from .common import get_paths
from .test_plugin_flask import api as flask_api


def test_plugin_spec():
    api = flask_api
    assert api.spec["tags"] == [{"name": tag} for tag in ("test", "health", "api")]

    assert get_paths(api.spec) == [
        "/api/file",
        "/api/group/{name}",
        "/api/user",
        "/api/user/{name}",
        "/ping",
        "/v1/api/file",
        "/v1/api/group/{name}",
        "/v1/api/user",
        "/v1/api/user/{name}",
        "/v1/ping",
    ]

    ping = api.spec["paths"]["/ping"]["get"]
    assert ping["tags"] == ["test", "health"]
    assert ping["parameters"][0]["in"] == "header"
    assert ping["summary"] == "summary"
    assert ping["description"] == "description"

    user = api.spec["paths"]["/api/user/{name}"]["post"]
    assert user["tags"] == ["api", "test"]
    assert (
        user["requestBody"]["content"]["application/json"]["schema"]["$ref"]
        == "#/components/schemas/JSON"
    )
    assert len(user["responses"]) == 3

    params = user["parameters"]
    for param in params:
        if param["in"] == "path":
            assert param["name"] == "name"
        elif param["in"] == "query":
            assert param["name"] == "order"



import pytest

from flask_pydantic_spec.utils import (
    parse_comments,
    parse_request,
    parse_params,
    parse_resp,
    has_model,
    parse_name,
)
from flask_pydantic_spec.spec import FlaskPydanticSpec
from flask_pydantic_spec.types import Response, Request, _parse_code

from .common import DemoModel


api = FlaskPydanticSpec()


def undecorated_func():
    """summary
    description"""
    pass


@api.validate(body=Request(model=DemoModel), resp=Response(HTTP_200=DemoModel))
def demo_func():
    """
    summary

    description"""
    pass


class DemoClass:
    @api.validate(query=DemoModel, resp=Response(HTTP_200=DemoModel))
    def demo_method(self):
        """summary
        description
        """
        pass


demo_class = DemoClass()


def test_comments():
    assert parse_comments(lambda x: x) == (None, None)
    assert parse_comments(undecorated_func) == ("summary", "description")
    assert parse_comments(demo_func) == ("summary", "description")
    assert parse_comments(demo_class.demo_method) == ("summary", "description")


def test_parse_code():
    with pytest.raises(TypeError):
        assert _parse_code(200) == 200

    assert _parse_code("200") is None
    assert _parse_code("HTTP_404") == "404"


def test_parse_name():
    assert parse_name(lambda x: x) == "<lambda>"
    assert parse_name(undecorated_func) == "undecorated_func"
    assert parse_name(demo_func) == "demo_func"
    assert parse_name(demo_class.demo_method) == "demo_method"


def test_has_model():
    assert not has_model(undecorated_func)
    assert has_model(demo_func)
    assert has_model(demo_class.demo_method)


def test_parse_resp():
    assert parse_resp(undecorated_func, 422) == {}
    assert parse_resp(demo_class.demo_method, 422) == {
        "200": {
            "description": "OK",
            "content": {"application/json": {"schema": {"$ref": "#/components/schemas/DemoModel"}}},
        },
        "422": {"description": "Validation Error"},
    }
    resp_spec = parse_resp(demo_func, 422)
    assert resp_spec["422"]["description"] == "Validation Error"
    assert (
        resp_spec["200"]["content"]["application/json"]["schema"]["$ref"]
        == "#/components/schemas/DemoModel"
    )

    resp_spec = parse_resp(demo_func, 400)
    assert "422" not in resp_spec
    assert resp_spec["400"]["description"] == "Validation Error"


def test_parse_request():
    assert (
        parse_request(demo_func)["content"]["application/json"]["schema"]["$ref"]
        == "#/components/schemas/DemoModel"
    )
    assert parse_request(demo_class.demo_method) == {}


def test_parse_params():
    models = {"DemoModel": DemoModel.model_json_schema()}
    assert parse_params(demo_func, [], models) == []
    params = parse_params(demo_class.demo_method, [], models)
    assert len(params) == 3
    assert params[0] == {
        "name": "uid",
        "in": "query",
        "required": True,
        "schema": {
            "title": "Uid",
            "type": "integer",
        },
    }



import inspect
import json
import logging
import re
from json import JSONDecodeError

from typing import (
    Callable,
    Mapping,
    Any,
    Tuple,
    Optional,
    List,
    Dict,
    Iterable,
    Type,
)

from werkzeug.datastructures import MultiDict
from pydantic import BaseModel, v1
from werkzeug.routing import Rule

from .constants import OPENAPI_SCHEMA_TEMPLATE
from .types import BaseModelT, BaseModelUnion, Response, RequestBase, Request

logger = logging.getLogger(__name__)

VALID_NAME_REGEX = re.compile(r"[^a-zA-Z0-9._-]")


def get_model_name(model: Type[BaseModelUnion]) -> str:
    """Gets the name of a model name as an OpenAPI 3.1 compatible name

    Replaces any non standard characters in a string with `_`

    >>> format_model_name("AB[C[D]]")
    AB_C_D__
    """
    return VALID_NAME_REGEX.sub("_", model.__name__)


def get_model_schema(model: Type[BaseModelUnion]) -> Dict[str, Any]:
    if issubclass(model, BaseModel):
        return model.model_json_schema(ref_template=OPENAPI_SCHEMA_TEMPLATE)
    elif issubclass(model, v1.BaseModel):
        return model.schema(ref_template=OPENAPI_SCHEMA_TEMPLATE)
    else:
        raise ValueError(f"Unsupported model type: {type(model)}")


def load_model_schema(model: Type[BaseModelT], data: Any) -> BaseModelT:
    if issubclass(model, BaseModel):
        return model.model_validate(data)  # type: ignore[return-value]
    elif issubclass(model, v1.BaseModel):
        return model.parse_obj(data)  # type: ignore[return-value]
    else:
        raise ValueError(f"Unsupported model type: {type(model)}")


def parse_comments(func: Callable) -> Tuple[Optional[str], Optional[str]]:
    """
    parse function comments

    First line of comments will be saved as summary, and the rest
    will be saved as description.
    """
    doc = inspect.getdoc(func)
    if doc is None:
        return None, None
    docs = doc.split("\n", 1)
    if len(docs) == 1:
        return docs[0], None
    return docs[0], docs[1].strip()


def parse_request(func: Callable) -> Mapping[str, Any]:
    """
    Generate spec from body parameter on the view function validation decorator
    """
    if hasattr(func, "body"):
        request_body = getattr(func, "body")
        if isinstance(request_body, RequestBase):
            result: Mapping[str, Any] = request_body.generate_spec()
        elif issubclass(request_body, (BaseModel, v1.BaseModel)):
            result = Request(request_body).generate_spec()
        else:
            result = {}
        return result
    return {}


def parse_params(
    func: Callable,
    params: List[Mapping[str, Any]],
    models: Mapping[str, Any],
) -> List[Mapping[str, Any]]:
    """
    get spec for (query, headers, cookies)
    """
    if hasattr(func, "query"):
        model_name = getattr(func, "query").__name__
        query = models.get(model_name)
        if query is not None:
            for name, schema in query["properties"].items():
                params.append(
                    {
                        "name": name,
                        "in": "query",
                        "schema": schema,
                        "required": name in query.get("required", []),
                    }
                )

    if hasattr(func, "headers"):
        model_name = getattr(func, "headers").__name__
        headers = models.get(model_name)
        if headers is not None:
            for name, schema in headers["properties"].items():
                params.append(
                    {
                        "name": name,
                        "in": "header",
                        "schema": schema,
                        "required": name in headers.get("required", []),
                    }
                )

    if hasattr(func, "cookies"):
        model_name = getattr(func, "cookies").__name__
        cookies = models.get(model_name)
        if cookies is not None:
            for name, schema in cookies["properties"].items():
                params.append(
                    {
                        "name": name,
                        "in": "cookie",
                        "schema": schema,
                        "required": name in cookies.get("required", []),
                    }
                )

    return params


def parse_resp(func: Callable, code: int) -> Mapping[str, Mapping[str, Any]]:
    """
    get the response spec

    If this function does not have explicit ``resp`` but have other models,
    a ``Validation Error`` will be append to the response spec. Since
    this may be triggered in the validation step.
    """
    responses: Dict[str, Any] = {}
    if hasattr(func, "resp"):
        response = getattr(func, "resp")
        if response:
            responses = response.generate_spec()

    if str(code) not in responses and has_model(func):
        responses[str(code)] = {"description": "Validation Error"}

    return responses


def has_model(func: Callable) -> bool:
    """
    return True if this function have ``pydantic.BaseModel``
    """
    if any(hasattr(func, x) for x in ("query", "json", "headers")):
        return True

    if hasattr(func, "resp") and getattr(func, "resp").has_model():
        return True

    return False


def parse_name(func: Callable) -> str:
    """
    the func can be

        * undecorated functions
        * decorated functions
        * decorated class methods
    """
    return func.__name__


def default_before_handler(
    req: Request, resp: Response, req_validation_error: Any, instance: BaseModel
) -> None:
    """
    default handler called before the endpoint function after the request validation

    :param req: request provided by the web framework
    :param resp: response generated by Flask_Pydantic_Spec that will be returned
        if the validation error is not None
    :param req_validation_error: request validation error
    :param instance: class instance if the endpoint function is a class method
    """
    if req_validation_error:
        logger.info(
            "Validation Error",
            extra={
                "spectree_model": req_validation_error.__class__.__name__,
                "spectree_validation": req_validation_error.errors(),
            },
        )


def default_after_handler(
    req: Request, resp: Response, resp_validation_error: Any, instance: BaseModel
) -> None:
    """
    default handler called after the response validation

    :param req: request provided by the web framework
    :param resp: response from the endpoint function (if there is no validation error)
        or response validation error
    :param resp_validation_error: response validation error
    :param instance: class instance if the endpoint function is a class method
    """
    if resp_validation_error:
        logger.info(
            "500 Response Validation Error",
            extra={
                "spectree_model": resp_validation_error.__class__.__name__,
                "spectree_validation": resp_validation_error.errors(),
            },
        )


def parse_multi_dict(input: MultiDict) -> Dict[str, Any]:
    result = {}
    for key, value in input.to_dict(flat=False).items():
        if len(value) == 1:
            try:
                value_to_use = json.loads(value[0])
            except (TypeError, JSONDecodeError):
                value_to_use = value[0]
        else:
            value_to_use = value
        result[key] = value_to_use
    return result


RE_PARSE_RULE = re.compile(
    r"""
    (?P<static>[^<]*)                           # static rule data
    <
    (?:
        (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
        (?:\((?P<args>.*?)\))?                  # converter arguments
        \:                                      # variable delimiter
    )?
    (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)        # variable name
    >
    """,
    re.VERBOSE,
)


def parse_rule(rule: Rule) -> Iterable[Tuple[Optional[str], Optional[str], str]]:
    """
    Parse a rule and return it as generator. Each iteration yields tuples in the form
    ``(converter, arguments, variable)``.
    If the converter is `None` it's a static url part, otherwise it's a dynamic one.
    Note: This originally lived in werkzeug.routing.parse_rule until it was
    removed in werkzeug 2.2.0.
    TODO - cgearing - do we really need this?
    """
    rule_str = str(rule)
    pos = 0
    end = len(rule_str)
    do_match = RE_PARSE_RULE.match
    used_names = set()
    while pos < end:
        m = do_match(rule_str, pos)
        if m is None:
            break
        data = m.groupdict()
        if data["static"]:
            yield None, None, data["static"]
        variable = data["variable"]
        converter = data["converter"] or "default"
        if variable in used_names:
            raise ValueError(f"variable name {variable!r} used twice.")
        used_names.add(variable)
        yield converter, data["args"] or None, variable
        pos = m.end()
    if pos < end:
        remaining = rule_str[pos:]
        if ">" in remaining or "<" in remaining:
            raise ValueError(f"malformed url rule: {rule_str!r}")
        yield None, None, remaining



import pytest

from flask_pydantic_spec.types import (
    DEFAULT_CODE_DESC,
    Response,
    FileResponse,
    Request,
    MultipartFormRequest,
)

from .common import DemoModel, DemoModelV1


class NormalClass:
    pass


@pytest.fixture(
    params=[
        pytest.param(DemoModel, id="v2"),
        pytest.param(DemoModelV1, id="v1"),
    ]
)
def model_cls(request):
    return request.param


def test_init_response(model_cls):
    for args, kwargs in [
        ([200], {}),
        (["HTTP_110"], {}),
        ([], {"HTTP_200": NormalClass}),
    ]:
        with pytest.raises(AssertionError):
            Response(*args, **kwargs)

    resp = Response("HTTP_200", HTTP_201=model_cls)
    assert resp.has_model()
    assert resp.find_model(201) == model_cls
    assert model_cls in resp.models

    resp = Response(HTTP_200=None, HTTP_403=model_cls)
    assert resp.has_model()
    assert resp.find_model(403) == model_cls
    assert resp.find_model(200) is None
    assert model_cls in resp.models

    assert not Response().has_model()


def test_response_spec(model_cls):
    resp = Response("HTTP_200", HTTP_201=model_cls)
    spec = resp.generate_spec()
    assert spec["200"]["description"] == DEFAULT_CODE_DESC["HTTP_200"]
    assert spec["201"]["description"] == DEFAULT_CODE_DESC["HTTP_201"]
    assert (
        spec["201"]["content"]["application/json"]["schema"]["$ref"].split("/")[-1]
        == model_cls.__name__
    )

    assert spec.get(200) is None
    assert spec.get(404) is None


def test_file_response_spec():
    octet_resp = FileResponse()
    spec = octet_resp.generate_spec()
    assert spec["200"]["description"] == DEFAULT_CODE_DESC["HTTP_200"]
    assert spec["404"]["description"] == DEFAULT_CODE_DESC["HTTP_404"]

    assert spec["200"]["content"]["application/octet-stream"]["schema"]["format"] == "binary"
    assert spec["200"]["content"]["application/octet-stream"]["schema"]["type"] == "string"

    pdf_resp = FileResponse("application/pdf")
    pdf_spec = pdf_resp.generate_spec()
    assert pdf_spec["200"]["description"] == DEFAULT_CODE_DESC["HTTP_200"]
    assert pdf_spec["404"]["description"] == DEFAULT_CODE_DESC["HTTP_404"]

    assert pdf_spec["200"]["content"]["application/pdf"]["schema"]["format"] == "binary"
    assert pdf_spec["200"]["content"]["application/pdf"]["schema"]["type"] == "string"


def test_file_request_spec():
    file_request = Request(content_type="application/octet-stream")
    spec = file_request.generate_spec()
    assert spec["content"] == {
        "application/octet-stream": {"schema": {"type": "string", "format": "binary"}}
    }


def test_multipart_form_spec(model_cls):
    form = MultipartFormRequest(model_cls, "fileName")
    spec = form.generate_spec()
    assert spec["content"] == {
        "multipart/form-data": {
            "schema": {
                "type": "object",
                "properties": {
                    "uid": {
                        "type": "integer",
                        "title": "Uid",
                    },
                    "limit": {"type": "integer", "title": "Limit"},
                    "name": {"type": "string", "title": "Name"},
                    "fileName": {"type": "string", "format": "binary"},
                },
            }
        }
    }


def test_multipart_form_no_model():
    form = MultipartFormRequest()
    spec = form.generate_spec()
    assert spec["content"] == {
        "multipart/form-data": {
            "schema": {
                "type": "object",
                "properties": {
                    "file": {"type": "string", "format": "binary"},
                },
            }
        }
    }



import logging

from .types import Response, Request, MultipartFormRequest, FileResponse
from .spec import FlaskPydanticSpec

__all__ = [
    "FlaskPydanticSpec",
    "Response",
    "Request",
    "MultipartFormRequest",
    "FileResponse",
]

# setup library logging
logging.getLogger(__name__).addHandler(logging.NullHandler())



from enum import Enum
import re
from typing import Any, Optional, List

import pytest
from flask import Flask
from openapi_spec_validator import OpenAPIV31SpecValidator
from pydantic import BaseModel, StrictFloat, Field, RootModel
from pydantic import v1

from flask_pydantic_spec import Response
from flask_pydantic_spec.types import FileResponse, Request, MultipartFormRequest
from flask_pydantic_spec import FlaskPydanticSpec
from flask_pydantic_spec.config import Config
from flask_pydantic_spec.utils import get_model_name

from .common import ExampleConverter, UnknownConverter


class ExampleModel(BaseModel):
    name: str = Field(json_schema_extra={"strip_whitespace": True})
    age: int
    height: StrictFloat


class TypeEnum(str, Enum):
    foo = "foo"
    bar = "bar"


class ExampleQuery(BaseModel):
    query: str
    type: Optional[TypeEnum] = None


class ExampleNestedList(RootModel):
    root: List[ExampleModel]


class ExampleNestedModel(BaseModel):
    example: ExampleModel


class ExampleDeepNestedModel(BaseModel):
    data: List["ExampleModel"]


class ExampleV1Model(v1.BaseModel):
    name: str = v1.Field(strip_whitespace=True)
    age: int
    height: StrictFloat


class ExampleV1Query(v1.BaseModel):
    query: str
    type: Optional[TypeEnum] = None


class ExampleV1NestedList(v1.BaseModel):
    __root__: List[ExampleV1Model]


class ExampleV1NestedModel(v1.BaseModel):
    example: ExampleV1Model


class ExampleV1DeepNestedModel(v1.BaseModel):
    data: List["ExampleV1Model"]


def backend_app():
    return [
        ("flask", Flask(__name__)),
    ]


@pytest.fixture
def empty_app():
    return Flask(__name__)


@pytest.fixture
def name():
    return "flask"


@pytest.fixture
def api(name) -> FlaskPydanticSpec:
    return FlaskPydanticSpec(
        name,
        tags=[{"name": "lone", "description": "a lone api"}],
        validation_error_code=400,
    )


def test_spectree_init():
    spec = FlaskPydanticSpec(path="docs")
    conf = Config()

    assert spec.config.TITLE == conf.TITLE
    assert spec.config.PATH == "docs"


def test_register(name, empty_app):
    api = FlaskPydanticSpec(name)
    api.register(empty_app)


def test_spec_generate(name, empty_app):
    api = FlaskPydanticSpec(
        name,
        app=empty_app,
        title=f"{name}",
        info={"title": "override", "description": "api level description"},
        tags=[{"name": "lone", "description": "a lone api"}],
    )
    spec = api.spec

    assert spec["info"]["title"] == name
    assert spec["info"]["description"] == "api level description"
    assert spec["paths"] == {}
    assert spec["tags"] == []


@pytest.fixture
def app(api: FlaskPydanticSpec) -> Flask:
    app = Flask(__name__)
    app.url_map.converters["example"] = ExampleConverter
    app.url_map.converters["unknown"] = UnknownConverter

    @app.get("/foo")
    @api.validate(resp=Response(HTTP_200=ExampleModel))
    def foo():
        pass

    @app.post("/lone")
    @api.validate(
        body=Request(ExampleModel),
        resp=Response(HTTP_200=ExampleNestedList, HTTP_400=ExampleNestedModel),
        tags=["lone"],
        deprecated=True,
    )
    def lone_post():
        pass

    @app.patch("/lone")
    @api.validate(
        body=Request(ExampleModel),
        resp=Response(HTTP_200=List[ExampleModel], HTTP_400=ExampleNestedModel),
        tags=["lone"],
    )
    def lone_patch():
        pass

    @app.get("/query")
    @api.validate(query=ExampleQuery, resp=Response(HTTP_200=List[ExampleModel]))
    def get_query():
        pass

    @app.get("/file")
    @api.validate(resp=FileResponse())
    def get_file():
        pass

    @app.post("/file")
    @api.validate(
        body=Request(content_type="application/octet-stream"),
        resp=FileResponse(),
    )
    def post_file():
        pass

    @app.post("/multipart-file")
    @api.validate(body=MultipartFormRequest(ExampleModel), resp=Response(HTTP_200=ExampleModel))
    def post_multipart_form():
        pass

    @app.get("/enum/<example:example>")
    @api.validate(resp=Response(HTTP_200=ExampleModel))
    def get_enum(example):
        pass

    @app.get("/v1/foo")
    @api.validate(resp=Response(HTTP_200=ExampleV1Model))
    def foo_v1():
        pass

    @app.post("/v1/lone")
    @api.validate(
        body=ExampleV1Model,
        resp=Response(HTTP_200=ExampleV1NestedList, HTTP_400=ExampleV1NestedModel),
        tags=["lone"],
        deprecated=True,
    )
    def lone_post_v1():
        pass

    @app.patch("/v1/lone")
    @api.validate(
        body=Request(ExampleV1Model),
        resp=Response(HTTP_200=List[ExampleV1Model], HTTP_400=ExampleV1NestedModel),
        tags=["lone"],
    )
    def lone_patch_v1():
        pass

    @app.get("/v1/query")
    @api.validate(query=ExampleV1Query, resp=Response(HTTP_200=List[ExampleV1Model]))
    def get_query_v1():
        pass

    @app.post("/v1/multipart-file")
    @api.validate(body=MultipartFormRequest(ExampleV1Model), resp=Response(HTTP_200=ExampleV1Model))
    def post_multipart_form_v1():
        pass

    @app.get("/v1/enum/<example:example>")
    @api.validate(resp=Response(HTTP_200=ExampleV1Model))
    def get_enum_v1(example):
        pass

    return app


def test_two_endpoints_with_the_same_path(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec

    http_methods = list(spec["paths"]["/lone"].keys())
    http_methods.sort()
    assert http_methods == ["patch", "post"]


def test_valid_openapi_spec(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec

    assert OpenAPIV31SpecValidator(spec).is_valid()


def test_openapi_tags(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec

    assert spec["tags"][0]["name"] == "lone"
    assert spec["tags"][0]["description"] == "a lone api"


def test_openapi_deprecated(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec

    assert spec["paths"]["/lone"]["post"]["deprecated"] is True
    assert "deprecated" not in spec["paths"]["/lone"]["patch"]


def test_flat_array_schemas(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec
    assert spec["components"]["schemas"][get_model_name(ExampleNestedList)].get("items") is not None


@pytest.mark.parametrize(
    ("route", "schema"),
    [
        pytest.param(
            "/convert/<any(a, b, c):example>",
            {"type": "string", "enum": ["a", "b", "c"]},
            id="any",
        ),
        pytest.param(
            "/convert/<int(min=1, max=5):example>",
            {"type": "integer", "format": "int32", "minimum": 1, "maximum": 5},
            id="int",
        ),
        pytest.param(
            "/convert/<uuid:example>",
            {"type": "string", "format": "uuid"},
            id="uuid",
        ),
        pytest.param(
            "/convert/<float:example>",
            {"type": "number", "format": "float"},
            id="float",
        ),
        pytest.param(
            "/convert/<path:example>",
            {"type": "string", "format": "path"},
            id="path",
        ),
        pytest.param(
            "/convert/<string(length=5):example>",
            {"type": "string", "length": 5},
            id="string-with-length",
        ),
        pytest.param(
            "/convert/<string(maxlength=5):example>",
            {"type": "string", "maxLength": 5},
            id="string-with-max-length",
        ),
        pytest.param(
            "/convert/<unknown:example>",
            {"type": "string"},
            id="custom-unknown",
        ),
        pytest.param(
            "/convert/<example:example>",
            {"type": "string", "enum": ["one", "two"]},
            id="custom-enum",
        ),
    ],
)
def test_url_converters(route, schema, app: Flask, api: FlaskPydanticSpec):
    @app.get(route)
    @api.validate(resp=Response(HTTP_200=None))
    def get_with_converter(example):
        pass

    api.register(app)

    spec = api.spec

    spec_route = re.sub(r"<.*:(.*)>", r"{\1}", route)

    assert spec["paths"][spec_route]["get"]["parameters"][0]["schema"] == schema


def test_flat_array_schema_from_python_list_type(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec

    schema_spec = spec["paths"]["/lone"]["patch"]["responses"]["200"]["content"][
        "application/json"
    ]["schema"]

    assert (
        schema_spec["type"] == "array"
        and schema_spec["items"]["$ref"] == "#/components/schemas/ExampleModel"
    )


# TODO: add tests for v1 schemas
def strip_v1(data: Any) -> Any:
    if isinstance(data, dict):
        return {k: strip_v1(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [strip_v1(item) for item in data]
    elif isinstance(data, str):
        return data.replace("V1", "").replace("_v1", "")
    else:
        return data


@pytest.mark.parametrize(
    ("route", "method"),
    [
        ("/foo", "get"),
        ("/lone", "post"),
        ("/lone", "patch"),
        ("/multipart-file", "post"),
        ("/enum/{example}", "get"),
    ],
)
def test_v1_routes_match_v2(app: Flask, api: FlaskPydanticSpec, route: str, method: str):
    api.register(app)
    spec = api.spec
    v1_route = f"/v1{route}"

    v1_spec = spec["paths"][v1_route][method]
    v2_spec = spec["paths"][route][method]

    assert strip_v1(v1_spec) == v2_spec


@pytest.mark.parametrize(
    ("route", "method"),
    [
        ("/query", "get"),
    ],
)
def test_v1_routes_with_nullable_match(app: Flask, api: FlaskPydanticSpec, route: str, method: str):
    api.register(app)
    spec = api.spec
    v1_route = f"/v1{route}"

    v1_spec = spec["paths"][v1_route][method]
    v2_spec = spec["paths"][route][method]

    v1_query_type = v1_spec["parameters"][1].pop("schema")
    v2_query_type = v2_spec["parameters"][1].pop("schema")

    assert strip_v1(v1_spec) == v2_spec
    # Pydantic v1 was incorrectly implemented
    assert v1_query_type == {"$ref": "#/components/schemas/TypeEnum"}
    assert v2_query_type == {
        "anyOf": [{"$ref": "#/components/schemas/TypeEnum"}, {"type": "null"}],
        "default": None,
    }


def test_v1_route_request_bodies(app: Flask, api: FlaskPydanticSpec):
    api.register(app)
    spec = api.spec

    v1_spec = spec["paths"]["/v1/lone"]["post"]

    assert v1_spec["requestBody"] == {
        "content": {"application/json": {"schema": {"$ref": "#/components/schemas/ExampleV1Model"}}}
    }
